package com.jkcredit.query.record.service.impl;

import com.alibaba.fastjson.JSON;
import com.ciphergateway.ciphersuite.CipherSuiteException;
import com.ciphergateway.ciphersuite.CipherSuiteMacException;
import com.ciphergateway.ciphersuite.CipherSuiteUtils;
import com.jkcredit.query.record.constant.CommonConstant;
import com.jkcredit.query.record.model.*;
import com.jkcredit.query.record.service.QueryRecordService;
import com.jkcredit.query.record.util.IndexUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.client.core.CountRequest;
import org.elasticsearch.client.core.CountResponse;
import org.elasticsearch.client.indices.GetIndexRequest;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.SearchHits;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.xml.bind.DatatypeConverter;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * @author xusonglin
 * @version V1.0
 **/
@Service
@Slf4j
public class QueryRecordServiceImpl implements QueryRecordService {
    @Value("${recordsEncrypt.keyId}")
    private String keyId;
    @Value("${recordsEncrypt.metadata}")
    private String metadata;
    @Value("${recordsEncrypt.ivStr}")
    private String ivStr;
    @Value("${recordsEncrypt.algorithm}")
    private String algorithm;

    @Autowired
    private RestHighLevelClient esRestClient;
    private static SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();

    /**
     * 根据月份，车牌号查询
     *
     * @param param 车牌号 月份
     * @return 结果
     */
    @Override
    public CommonResponseObject monthRecords(MonthRecordsRequestParam param) {
        if (StringUtils.isBlank(param.getPlateNumber()) || StringUtils.isBlank(param.getMonth())) {
            return new CommonResponseObject(CommonConstant.PARAM_ERROR_CODE, CommonConstant.PARAM_ERROR);
        }
        String[] indices = {CommonConstant.ES_INDEX + param.getMonth()};
        List<MonthResult> results = find(param.getPlateNumber(), indices);
        return new CommonResponseObject(CommonConstant.SUCCESS_CODE, JSON.toJSONString(results));
    }

    /**
     * 根据车牌号，查询12个月的结果
     *
     * @param param 车牌号
     * @return 结果
     */
    @Override
    public CommonResponseObject yearRecords(YearRecordsRequestParam param) {
        if (StringUtils.isBlank(param.getPlateNumber())) {
            return new CommonResponseObject(CommonConstant.PARAM_ERROR_CODE, CommonConstant.PARAM_ERROR);
        }

        String[] indices = new String[12];
        LocalDate today = LocalDate.now();
        for (int i = 0; i < 12; i++) {
            LocalDate localDate = today.minusMonths(i);
            String month = localDate.toString().substring(0, 7).replace("-", "");
            indices[i] = CommonConstant.ES_INDEX + month;
        }
        List<MonthResult> results = find(param.getPlateNumber(), indices);
        return new CommonResponseObject(CommonConstant.SUCCESS_CODE, JSON.toJSONString(results));
    }

    /**
     * 查询索引中结果总数
     *
     * @param param 月份
     * @return 结果总数
     */
    @Override
    public CommonResponseObject countByMonth(RecordsCountRequestParam param) {
        if (StringUtils.isBlank(param.getMonth())) {
            return new CommonResponseObject(CommonConstant.PARAM_ERROR_CODE, CommonConstant.PARAM_ERROR);
        }

        String index = CommonConstant.ES_INDEX + param.getMonth();
        if (!isExistIndex(index)) {
            return new CommonResponseObject(CommonConstant.SUCCESS_CODE, "0");
        }
        //构建判断条件
        BoolQueryBuilder boolBuilder = QueryBuilders.boolQuery();
        searchSourceBuilder.timeout(new TimeValue(CommonConstant.OUT_TIME, TimeUnit.MILLISECONDS));
        searchSourceBuilder.query(boolBuilder);
        CountRequest countRequest = new CountRequest(index);
        CountResponse countResponse;
        long count = 0L;
        try {
            countResponse = esRestClient.count(countRequest, RequestOptions.DEFAULT);
            count = countResponse != null ? countResponse.getCount() : 0;
        } catch (Exception e) {
            log.error("queryRecordCountException:", e);
        }
        return new CommonResponseObject(CommonConstant.SUCCESS_CODE, String.valueOf(count));
    }

    /**
     * 根据车牌号，查询多月结果（小于12个月）
     *
     * @param param 车牌号 起始月份 结束月份
     * @return 结果
     */
    @Override
    public CommonResponseObject multipleMonthRecords(MultipleMonthRecordsRequestParam param) {
        if (StringUtils.isBlank(param.getStartMonth())
                || StringUtils.isBlank(param.getEndMonth())
                || StringUtils.isBlank(param.getPlateNumber())) {
            return new CommonResponseObject(CommonConstant.PARAM_ERROR_CODE, CommonConstant.PARAM_ERROR);
        }

        List<String> monthList = getMonthBetween(param.getStartMonth(), param.getEndMonth());
        if (monthList.size() > 12 || monthList.size() == 0) {
            return new CommonResponseObject(CommonConstant.MONTH_ERROR_CODE, CommonConstant.MONTH_ERROR);
        }
        String[] indices = new String[monthList.size()];
        for (int i = 0; i < monthList.size(); i++) {
            indices[i] = CommonConstant.ES_INDEX + monthList.get(i);
        }
        List<MonthResult> results = find(param.getPlateNumber(), indices);
        return new CommonResponseObject(CommonConstant.SUCCESS_CODE, JSON.toJSONString(results));
    }

    /**
     * 根据车牌号，索引查结果
     *
     * @param plateNumber 车牌号
     * @param indices     索引
     * @return 结果
     */
    private List<MonthResult> find(String plateNumber, String[] indices) {
        // 打印入参出参日志
        LogObject logObject = new LogObject();
        logObject.setPlateNumber(plateNumber);
        logObject.setIndices(indices);

        // 结果集合
        List<MonthResult> results = new ArrayList<>();

        // 判断索引是否存在
        List<String> existIndices = new ArrayList<>();
        for (String index : indices) {
            if (isExistIndex(index)) {
                existIndices.add(index);
            }
        }
        // 索引不存在，返回空结果
        if (existIndices.size() == 0) {
            logObject.setSuccess(false);
            logObject.setMessage(CommonConstant.INDEX_NOT_FIND);
            log.info(JSON.toJSONString(logObject));
            return results;
        }
        // 车牌号加密
        String encryptPlateNumber = vehicleIdEncrypt(plateNumber);
        if (StringUtils.isBlank(encryptPlateNumber)) {
            logObject.setSuccess(false);
            logObject.setMessage(CommonConstant.ENCRYPT_ERROR);
            log.info(JSON.toJSONString(logObject));
            log.error("encryptPlateNumberError:{}", plateNumber);
            return results;
        }
        // 设置es的索引index
        SearchRequest searchRequest = new SearchRequest();
        searchRequest.indices(existIndices.toArray(new String[0]));
        // bool工厂，相当于sql中where条件，where之后的条件都写在这里
        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
        queryBuilder.must(QueryBuilders.matchQuery("vehicleid.keyword", encryptPlateNumber));
        // 将where条件放入工厂
        searchSourceBuilder.timeout(new TimeValue(CommonConstant.OUT_TIME, TimeUnit.MILLISECONDS));
        searchSourceBuilder.query(queryBuilder);
        searchSourceBuilder.size(CommonConstant.MAX_RESULT_SIZE);
        searchRequest.source(searchSourceBuilder);
        SearchResponse searchResponse = null;

        long startTime = System.currentTimeMillis();
        try {
            // 执行请求
            searchResponse = esRestClient.search(searchRequest, RequestOptions.DEFAULT);
        } catch (IOException e) {
            log.error("queryRecordException:", e);
        }
        long endTime = System.currentTimeMillis();

        if (searchResponse == null) {
            logObject.setSuccess(false);
            logObject.setMessage(CommonConstant.QUERY_ERROR);
            log.info(JSON.toJSONString(logObject));
            return results;
        }
        // 返回查询到的具体数据
        SearchHits searchHits = searchResponse.getHits();
        for (SearchHit hit : searchHits.getHits()) {
            MonthResult monthResult = JSON.toJavaObject(JSON.parseObject(hit.getSourceAsString()), MonthResult.class);
            results.add(monthResult);
        }

        logObject.setSuccess(true);
        logObject.setMessage(CommonConstant.QUERY_SUCCESS);
        logObject.setResult(results);
        logObject.setQueryRecordSuccess(endTime - startTime);
        log.info(JSON.toJSONString(logObject));
        return results;
    }

    /**
     * 判断索引是否存在
     *
     * @param index 索引
     * @return 结果
     */
    private boolean isExistIndex(String index) {
        boolean isExistIndex = false;
        String indexInMap = IndexUtil.INSTANCE.getByIndex(index);
        if (StringUtils.isBlank(indexInMap)) {
            GetIndexRequest getIndexRequest = new GetIndexRequest(index);
            try {
                isExistIndex = esRestClient.indices().exists(getIndexRequest, RequestOptions.DEFAULT);
            } catch (Exception e) {
                e.printStackTrace();
            }
            if (isExistIndex) {
                IndexUtil.INSTANCE.put(index, String.valueOf(System.currentTimeMillis()));
            }
        } else {
            isExistIndex = true;
        }

        return isExistIndex;
    }

    /**
     * 获取月份区间
     *
     * @param minDate 开始月
     * @param maxDate 结束月
     * @return 月份集合
     */
    private List<String> getMonthBetween(String minDate, String maxDate) {
        ArrayList<String> result = new ArrayList<String>();
        try {
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMM");//格式化为年月

            Calendar min = Calendar.getInstance();
            Calendar max = Calendar.getInstance();
            min.setTime(sdf.parse(minDate));
            min.set(min.get(Calendar.YEAR), min.get(Calendar.MONTH), 1);

            max.setTime(sdf.parse(maxDate));
            max.set(max.get(Calendar.YEAR), max.get(Calendar.MONTH), 2);

            Calendar curr = min;
            while (curr.before(max)) {
                result.add(sdf.format(curr.getTime()));
                curr.add(Calendar.MONTH, 1);
            }
        } catch (ParseException e) {
            e.printStackTrace();
        }
        return result;
    }

    /**
     * 对vehicle_id字段进行加密
     *
     * @param vehicleId 待加密vehicleId
     * @return 加密后vehicleId
     */
    private String vehicleIdEncrypt(String vehicleId) {
        try {
            byte[] plainData = vehicleId.getBytes("UTF-8");
            byte[] iv = ivStr.getBytes("UTF-8");
            // 加密
            byte[] cipherData = CipherSuiteUtils.encrypt(plainData, algorithm, keyId, metadata, iv);
            return DatatypeConverter.printHexBinary(cipherData);
        } catch (UnsupportedEncodingException ue) {
            log.error("UnsupportedEncodingException:", ue);
            return "";
        } catch (CipherSuiteException cse) {
            log.error("CipherSuiteException:", cse);
            return "";
        } catch (CipherSuiteMacException csme) {
            log.error("CipherSuiteMacException:", csme);
            return "";
        } catch (Exception e) {
            log.error("Exception:", e);
            return "";
        } catch (Error error) {
            log.error("Error:", error);
            return "";
        }
    }
}
